"""
Cleans and tokenizes given text
"""
import re
from typing import Callable, List


def clean(text: str) -> str:
    """Cleans up the provided wikitext.
    Removes templates, tables, parser hooks, magic words, HTML tags and file embeds.
    Keeps links.
    """
    # basic formatting
    text = re.sub(r"'''?([^']+)'''?", "\\1", text)

    # templates
    # {{foo}}
    # {{foo|{{test}}|123}}
    while "{{" in text:
        start = text.find("{{")
        level = 1
        pos = start + 2

        while pos < len(text):
            # print('>' + text[pos:pos+2] + '<')

            if text[pos : pos + 2] == "{{":
                # nested template - enter next level
                level += 1
                pos += 1
            elif text[pos : pos + 2] == "}}":
                # nested template - leave this level
                pos += 1
                level -= 1

            # template is now completed
            if level == 0:
                # print(text, start, pos, text[start:pos+1])
                text = text[:start] + " " + text[pos + 1 :]
                break

            # check next character
            pos += 1

        # the template is not well balanced, leave the endless loop
        if level != 0:
            break

    # tables
    text = re.sub(r"{\|[^}]+\|}", "", text)  # {|foo..|}

    # headings
    text = re.sub(
        r"^=+\s?([^=]+)\s?=+",
        lambda matches: matches.group(1).strip(),
        text,
        flags=re.MULTILINE,
    )  # == a == -> a

    # files and other links with namespaces
    text = re.sub(r"\[\[[^:\]]+:[^\]]+\]\]", "", text)  # [[foo:b]] -> ''

    # local links
    text = re.sub(r"\[\[([^|\]]+)\]\]", "\\1", text)  # [[a]] -> a
    text = re.sub(r"\[\[[^|]+\|([^\]]+)\]\]", "\\1", text)  # [[a|b]] -> b

    text = text.replace("[[", "").replace("]]", "")

    # external links
    text = re.sub(
        r"\[http[^\s]+ ([^\]]+)\]", "\\1", text
    )  # [[http://example.com foo]] -> foo
    text = re.sub(r"https?://[^\s]+", "", text)  # remove http://example.com

    # lists
    text = re.sub(r"^\*+\s?", "", text, flags=re.MULTILINE)

    # parser hooks
    text = re.sub(r"<[^>]+>[^<]+</[^>]+>", " ", text)  # <ref>foo</ref>

    # HTML
    text = re.sub(r"<[^>]+/?>", " ", text)  # <br> / <br />
    text = text.replace("&nbsp;", " ")

    # magic words
    text = re.sub(r"__\w+__", "", text)  # __TOC__

    return text.strip()


def tokenize_filter(text: str) -> bool:
    """Used by tokenize function below. Ignores numbers when tokenizing cleaned up wikitext."""
    if text == "":
        return False

    if text[0].isdigit() and re.fullmatch(r"\d+", text):
        return False

    return True


def tokenize(text: str, filter_func: Callable = tokenize_filter) -> List[str]:
    """Tokenizes a given text. Usually you want to first pass it via clean() function."""
    # clean up the text
    text = re.sub(r'[?.,:;!()=+"]', " ", text)  # remove noise

    text = text.strip()

    # tokenize
    parts = re.split(r"[-–\s/|_&{}\xAD]", text)

    parts = filter(filter_func, parts)

    return list(parts)



"""
Class used to parse provided XML dump and emit a stream of articles content

https://gist.github.com/macbre/1543d945f5244c5c68681966f07e2d6c
"""
import logging
from typing import Generator

from xml import sax
from xml.sax.xmlreader import AttributesImpl

from .dumps import BaseDump
from .entry import DumpEntry


class DumpHandler(sax.ContentHandler):
    """
    SAX ContentHandler instance of Wikipedia context XML dumps

      <page>
        <title>Wikipedia:All pages by title</title>
        <ns>4</ns>
        <id>4</id>
        <restrictions>sysop</restrictions>
        <revision>
          <id>6498</id>
          <timestamp>2004-05-25T02:19:28Z</timestamp>
          <contributor>
            <ip>82.68.206.22</ip>
          </contributor>
          <model>wikitext</model>
          <format>text/x-wiki</format>
          <text xml:space="preserve">foo bar</text>
          <sha1>edilrfbz8rbsr9mvl8dflarkw56ojy5</sha1>
        </revision>
      </page>
    """

    # https://www.mediawiki.org/wiki/Help:Export#Export_format
    # https://docs.python.org/3.6/library/xml.sax.handler.html#xml.sax.handler.ContentHandler

    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(self.__class__.__name__)

        self.entries_batch = []
        self.entries_count = 0

        # attributes from <mediawiki> root XML tag
        self.metadata = None

        # nodes from <siteinfo> XML tag
        self.siteinfo = {}
        self.base_url = None

        # parser state, are we inside <page> or <revision> tag?
        self.in_siteinfo = False
        self.in_page = False
        self.in_revision = False
        self.in_contributor = False

        # concatenated content of tags
        self.tag_content = ""

        # currently parsed page dump
        self.current_title = ""
        self.current_namespace = 0
        self.current_page_id = 0
        self.current_revision_id = 0
        self.current_revision_timestamp = 0
        self.current_content = ""
        self.current_contributor = None

    def reset_state(self):
        """
        Reset page information
        """
        self.in_siteinfo = False
        self.in_page = False
        self.in_revision = False
        self.in_contributor = False

        self.current_title = ""
        self.current_namespace = 0
        self.current_page_id = 0
        self.current_revision_id = 0
        self.current_revision_timestamp = 0
        self.current_content = ""
        self.current_contributor = None

    def startElement(self, name: str, attrs: AttributesImpl):
        """
        Run when a parser enters new element
        """
        # print('>', name, attrs)
        # self.logger.info('> startElement %s %s', name, attrs)

        if name == "siteinfo":
            self.in_siteinfo = True
        elif name == "page":
            self.in_page = True
        elif name == "revision":
            self.in_revision = True
        elif name == "contributor":
            self.in_contributor = True
        elif name == "mediawiki":
            self.metadata = dict(zip(attrs.keys(), attrs.values()))

        self.tag_content = ""

    # pylint: disable=too-many-branches
    def endElement(self, name: str):
        # print('<', name, self.tag_content)
        # self.logger.info('< endElement %s (%s)', name, self.tag_content)

        if name == "siteinfo":
            self.in_siteinfo = False
            self.reset_state()
            return
        if name == "page":
            self.in_page = False
            self.reset_state()
            return
        if name == "revision":
            self.in_revision = False

            # add next entry information
            self.logger.debug("Page #%d: %s", self.current_page_id, self.current_title)

            # build entry URL
            url = self.get_base_url() + self.current_title.replace(" ", "_")

            self.entries_batch.append(
                (
                    self.current_namespace,
                    self.current_page_id,
                    url,
                    self.current_title,
                    self.current_content,
                    self.current_revision_id,
                    self.current_revision_timestamp,
                    self.current_contributor,
                )
            )

            self.entries_count += 1
            return
        if name == "contributor":
            self.in_contributor = False
            return

        if self.in_contributor:
            if name == "username":
                self.current_contributor = self.tag_content
        elif self.in_revision:
            if name == "id":
                self.current_revision_id = int(self.tag_content)
            elif name == "timestamp":
                self.current_revision_timestamp = self.tag_content
            elif name == "text":
                self.current_content = self.tag_content
        elif self.in_page:
            if name == "title":
                self.current_title = self.tag_content
            elif name == "ns":
                self.current_namespace = int(self.tag_content)
            elif name == "id":
                self.current_page_id = int(self.tag_content)
        elif self.in_siteinfo:
            if name in ["dbname", "base", "generator"]:
                self.siteinfo[name] = self.tag_content

    def characters(self, content: str):
        # print('=', content)
        # self.logger.info('= characters %s', content)

        self.tag_content += content

    def get_entries(self) -> Generator[tuple, None, None]:
        """
        Used by DumpReader to yield pages as we parse the XML dump
        """
        for entry in self.entries_batch:
            yield entry

        self.entries_batch = []

    def get_entries_count(self) -> int:
        """
        :rtype: int
        """
        return self.entries_count

    def get_metadata(self) -> dict:
        """
        :rtype: dict|None
        """
        return self.metadata

    def get_siteinfo(self) -> dict:
        """
        :rtype: dict
        """
        return self.siteinfo

    def get_base_url(self) -> str:
        """
        :rtype: str
        """
        # e.g. base URL <https://pl.wikipedia.org/wiki/Wikipedia:Strona_g%C5%82%C3%B3wna>
        # will return <https://pl.wikipedia.org/wiki/>
        if self.base_url is None:
            main_page = str(self.siteinfo.get("base"))
            self.base_url = "/".join(main_page.split("/")[0:-1])
            self.base_url += "/"  # ends with slash

        return self.base_url


class DumpReader:
    """
    This class uses provided BaseDump instance to read and parse MediaWiki's XML dump
    """

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

        # https://docs.python.org/2/library/xml.etree.elementtree.html#parsing-xml
        self.handler = DumpHandler()

    @staticmethod
    def filter_by_namespace(namespace: int) -> bool:
        """
        :type namespace int
        :rtype bool
        """
        return isinstance(namespace, int)

    def read(self, dump: BaseDump) -> Generator[DumpEntry, None, None]:
        """Read a dump and emit DumpEntry objects"""
        self.logger.info("Parsing XML dump...")

        parser = sax.make_parser()
        parser.setContentHandler(self.handler)

        for chunk in dump.get_content():
            parser.feed(chunk)

            # yield pages as we go through XML stream
            for page in self.handler.get_entries():
                (
                    namespace,
                    page_id,
                    url,
                    title,
                    content,
                    revision_id,
                    revision_timestamp,
                    contributor,
                ) = page

                if self.filter_by_namespace(namespace):
                    if content == "":
                        # https://fo.wikipedia.org/wiki/Kjak:L%C3%ADvfr%C3%B8%C3%B0i
                        self.logger.warning("Page #%d: %s is empty", page_id, title)
                        continue

                    yield DumpEntry(
                        namespace,
                        page_id,
                        url,
                        title,
                        content,
                        revision_id,
                        revision_timestamp,
                        contributor,
                    )

        self.logger.info(
            "Parsing completed, entries found: %d", self.handler.get_entries_count()
        )

    def get_dump_language(self) -> str:
        """
        :rtype: str
        """
        return self.handler.get_metadata().get("xml:lang")

    def get_base_url(self) -> str:
        """
        :rtype: str
        """
        return self.handler.get_base_url()


class DumpReaderArticles(DumpReader):
    """
    Use this class to get article pages only
    """

    @staticmethod
    def filter_by_namespace(namespace: int) -> bool:
        """Process NS_MAIN articles only"""
        return namespace == 0



"""
A class representing dump entry
"""
from .utils import parse_date_string


# pylint: disable=too-many-instance-attributes
class DumpEntry:
    """
    An entry in XML dump
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        namespace: int,
        page_id: int,
        url: str,
        title: str,
        content: str,
        revision_id: int,
        timestamp: str,
        contributor: str = None,
    ):
        self.namespace = namespace
        self.page_id = page_id
        self.url = url
        self.title = title
        self.content = content
        self.revision_id = revision_id
        self.timestamp = timestamp
        self.contributor = contributor

    @property
    def unix_timestamp(self) -> float:
        """When was given article most recently edited"""
        return parse_date_string(self.timestamp).timestamp()

    def is_anon(self) -> bool:
        """Was this edit made by an anonymous contributor?"""
        return self.contributor is None

    def __repr__(self) -> str:
        contributor = self.contributor if not self.is_anon() else "Anonymous"
        date = parse_date_string(self.timestamp).isoformat()

        return f'<{self.__class__.__name__} "{self.title}" by {contributor} at {date}>'



"""
Utility functions
"""
from datetime import datetime, timezone


def parse_date_string(date: str) -> datetime:
    """Converts date as string (e.g. "2004-05-25T02:19:28Z") to UNIX timestamp (uses UTC, always)"""
    # https://docs.python.org/3.6/library/datetime.html#strftime-strptime-behavior
    # http://strftime.org/
    parsed = datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ")  # string parse time

    # now apply UTC timezone
    return parsed.replace(tzinfo=timezone.utc)



from setuptools import setup, find_packages

VERSION = "1.1.0"

# @see https://packaging.python.org/tutorials/packaging-projects/#creating-setup-py
with open("README.md", "r") as fh:
    long_description = fh.read()

# @see https://github.com/pypa/sampleproject/blob/master/setup.py
setup(
    name="mediawiki_dump",
    version=VERSION,
    author="Maciej Brencz",
    author_email="maciej.brencz@gmail.com",
    license="MIT",
    description="Python package for working with MediaWiki XML content dumps",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="dump fandom mediawiki wikipedia wikia",
    url="https://github.com/macbre/mediawiki-dump",
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Text Processing :: Markup :: XML",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.7",
    packages=find_packages(),
    extras_require={
        "dev": [
            "black==23.1.0",
            "coveralls==3.3.1",
            "pylint==2.17.0",
            "pytest==7.2.2",
            "pytest-cov==4.0.0",
            "responses==0.23.1",
        ]
    },
    install_requires=[
        "libarchive-c==4.0",
        "requests>=2.26.0",
        "mwclient>=0.10.1",
    ],
)



"""
CLasses that support fetching dumps
"""
import bz2
import logging

from typing import Generator, Iterator

from hashlib import md5
from os.path import isfile
from tempfile import gettempdir

from mwclient import Site
import requests
from requests.exceptions import HTTPError


class DumpError(Exception):
    """
    Generic exception class
    """


class BaseDump:
    """
    A generic dump class

    Wikipedia or Wikia dumps should have a separate class that customizes get_url() method
    """

    ARCHIVE_FORMAT = "bz2"

    def __init__(self, wiki, full_history=False):
        """
        :type wiki str
        :type full_history bool
        """
        self.wiki = wiki
        self.logger = logging.getLogger(self.__class__.__name__)

        self.http = requests.session()
        self.http.headers[
            "User-Agent"
        ] = "python-mediawiki-dump (+https://github.com/macbre/mediawiki-dump)"

        # do we want a full history or just the latest revisions?
        self.full_history = full_history

    def get_cache_filename(self, url):
        """
        Return a hashed filename of cache entry for a given URL

        :type url str
        :rtype: str
        """
        _hash = md5()
        _hash.update(url.encode("utf-8"))

        return f"mediawiki_dump_{_hash.hexdigest()}.{self.ARCHIVE_FORMAT}"

    def get_url(self):
        """
        :rtype: str
        """
        raise NotImplementedError("fetch method needs to be implemented")

    def fetch(self):
        """
        This method should be used internally only. Clients should call get_content().

        :rtype: _io.TextIOWrapper
        """
        url = self.get_url()

        cache_filename = f"{gettempdir()}/{self.get_cache_filename(url)}"
        self.logger.info("Checking %s cache file...", cache_filename)

        # check cache
        if not isfile(cache_filename):
            # fetch the resource
            self.logger.info("Fetching %s dump from <%s>...", self.wiki, url)
            response = self.http.get(url, stream=True)
            self.logger.info(
                "HTTP %s (%d kB will be fetched)",
                response.status_code,
                int(response.headers["content-length"]) / 1024,
            )

            # raise an exception and do not set a cache entry
            try:
                response.raise_for_status()
            except HTTPError as ex:
                self.logger.error("Failed to fetch a dump", exc_info=True)
                raise DumpError(
                    f"Failed to fetch a dump, request ended with HTTP {ex.response.status_code}"
                ) from ex

            # read the response as a stream and put it into cache file
            # http://docs.python-requests.org/en/master/user/advanced/#body-content-workflow
            #
            # before using a stream reading and parsing of Faroese dump made the words_from_dump.py
            # script took ~460 MB of memory, after the change - ~60 MB
            with open(cache_filename, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024):
                    if chunk:
                        file.write(chunk)

                response.close()
                self.logger.info("Cache set")
        else:
            self.logger.info("Reading from cache")

        # return a stream of compressed data from the cache file
        # pylint:disable=consider-using-with
        return open(cache_filename, "rb")

    def get_content(self) -> Generator[str, None, None]:
        """Yields processed pieces of content"""
        raise NotImplementedError("fetch method needs to be implemented")


class WikipediaDump(BaseDump):
    """
    Class for fetching Wikipedia dumps from https://dumps.wikimedia.org
    """

    def get_url(self):
        wiki = f"{self.wiki}wiki"
        version = "history" if self.full_history else "current"

        return (
            f"https://dumps.wikimedia.org/{wiki}/latest/"
            f"{wiki}-latest-pages-meta-{version}.xml.bz2"
        )

    def get_content(self) -> Generator[str, None, None]:
        """Yields processed pieces of content"""
        # https://docs.python.org/3.6/library/bz2.html#bz2.BZ2Decompressor
        decompressor = bz2.BZ2Decompressor()

        with self.fetch() as content:
            for chunk in content:
                yield decompressor.decompress(chunk)


class WikiaDump(BaseDump):
    """
    Class for fetching Wikia dumps

    https://community.wikia.com/wiki/Help:Database_download
    """

    ARCHIVE_FORMAT = "7z"

    def get_url(self) -> str:
        version = "full" if self.full_history else "current"

        # https://muppet.wikia.com/wiki/Special:Statistics
        return (
            f"https://s3.amazonaws.com/wikia_xml_dumps/"
            f"{self.wiki[0]}/{self.wiki[:2]}/{self.wiki}_pages_{version}.xml.7z"
        )

    def get_content(self) -> Generator[str, None, None]:
        """Yields processed pieces of content"""
        # https://github.com/Changaco/python-libarchive-c#usage
        try:
            # pylint:disable=import-outside-toplevel
            import libarchive
        except AttributeError as ex:
            # AttributeError: undefined symbol: archive_errno
            raise DumpError("Failed to import libarchive with 7zip support") from ex

        with self.fetch() as handler:
            with libarchive.file_reader(handler.name) as archive:
                for entry in archive:
                    for block in entry.get_blocks():
                        yield block


class LocalFileDump(BaseDump):
    """
    This class can be used to load locally stored XML dump file
    """

    def __init__(self, dump_file: str):
        super().__init__("")
        self.dump_file = dump_file

    def get_url(self):
        pass

    def get_content(self):
        """Yields processed pieces of content"""
        # pylint:disable=consider-using-with
        return open(self.dump_file, mode="rt", encoding="utf-8")


class StringDump(BaseDump):
    """
    This class can be used to load XML from a variable
    """

    def __init__(self, dump: str):
        super().__init__("")
        self.content = dump

    def get_url(self):
        pass

    def get_content(self):
        """Return the passed string"""
        return self.content


class LocalWikipediaDump(WikipediaDump):
    """
    This class can be used to load locally stored XML, bz2 compressed dump file
    """

    def __init__(self, dump_file: str):
        super().__init__("")
        self.dump_file = dump_file

    def get_url(self):
        pass

    def fetch(self):
        # pylint:disable=consider-using-with
        return open(self.dump_file, "rb")


class MediaWikiClientDump(BaseDump):
    """
    This class can be used to fetch "live" dumps from articles on any MediaWiki-powered site
    by using mwclient library
    """

    def __init__(self, site: Site, articles: Iterator[str]):
        """You must provide a mwclient.Site instance and a iterator that yields article names"""
        # https://mwclient.readthedocs.io/en/latest/index.html
        super().__init__("")

        self.site = site
        self.articles = list(articles)

    def get_url(self) -> str:
        return self.site.host  # e.g. vim.wikia.com

    def get_content(self) -> str:
        return self.fetch()

    def fetch(self) -> str:
        self.logger.info(
            "Fetching %d pages from %s wiki", len(self.articles), self.get_url()
        )

        # https://www.mediawiki.org/wiki/Manual:Parameters_to_Special:Export
        resp = self.site.raw_call(
            script="index",
            data={
                "title": "Special:Export",
                "curonly": "1",
                "pages": "\n".join(self.articles),
            },
            http_method="GET",
        )

        return resp
